# -*- coding: utf-8 -*-

import pickle

from contracts import describe_type
from . import logger
from .zc_debug_pickler import find_pickling_error
from .zc_safe_write import safe_read, safe_write

__all__ = [
    'safe_pickle_dump',
    'safe_pickle_load',
]

debug_pickling = False


def safe_pickle_dump(value, filename, protocol=pickle.HIGHEST_PROTOCOL,
                     **safe_write_options):
    with safe_write(filename, **safe_write_options) as f:
        try:
            pickle.dump(value, f, protocol)
        except KeyboardInterrupt:
            raise
        except BaseException:
            msg = 'Cannot pickle object of class %s' % describe_type(value)
            logger.error(msg)

            if debug_pickling:
                msg = find_pickling_error(value, protocol)
                logger.error(msg)
            raise


def safe_pickle_load(filename):
    # TODO: add debug check
    with safe_read(filename) as f:
        return pickle.load(f)
        # TODO: add pickling debug



# -*- coding: utf-8 -*-

import re
from typing import List, Union, Iterator

__all__ = [
    'expand_string',
    'get_wildcard_matches',
    'wildcard_to_regexp',
]


def flatten(seq):
    res = []
    for l in seq:
        res.extend(l)
    return res


def expand_string(x: Union[str, List[str]], options: List[str]) -> List[str]:
    if isinstance(x, list):
        return flatten(expand_string(y, options) for y in x)
    elif isinstance(x, str):
        x = x.strip()
        if ',' in x:
            splat = [_ for _ in x.split(',') if _]  # remove empty
            return flatten(expand_string(y, options) for y in splat)
        elif '*' in x:
            xx = expand_wildcard(x, options)
            expanded = list(xx)
            return expanded
        else:
            return [x]
    else:
        assert False


def wildcard_to_regexp(arg: str):
    """ Returns a regular expression from a shell wildcard expression. """
    return re.compile('\A' + arg.replace('*', '.*') + '\Z')


def has_wildcard(s: str) -> bool:
    return s.find('*') > -1


def expand_wildcard(wildcard: str, universe: List[str]) -> List[str]:
    '''
        Expands a wildcard expression against the given list.
        Raises ValueError if none found.

        :param wildcard: string with '*'
        :param universe: a list of strings
    '''
    if not has_wildcard(wildcard):
        msg = 'No wildcards in %r.' % wildcard
        raise ValueError(msg)

    matches = list(get_wildcard_matches(wildcard, universe))

    if not matches:
        msg = ('Could not find matches for pattern %r in %s.' %
               (wildcard, universe))
        raise ValueError(msg)

    return matches


def get_wildcard_matches(wildcard: str, universe: List[str]) -> Iterator[str]:
    '''
        Expands a wildcard expression against the given list.
        Yields a sequence of strings.

        :param wildcard: string with '*'
        :param universe: a list of strings
    '''
    regexp = wildcard_to_regexp(wildcard)
    for x in universe:
        if regexp.match(x):
            yield x



# -*- coding: utf-8 -*-
import os


def expand_all(x0):
    x = x0
    x = os.path.expanduser(x)
    x = os.path.expandvars(x)
    if '$' in x:
        msg = 'Cannot resolve all environment variables in %r.' % x0
        raise ValueError(msg)
    return x



import os

__all__ = [
    'mkdirs_thread_safe',
    'make_sure_dir_exists',
]

def mkdirs_thread_safe(dst):
    """Make directories leading to 'dst' if they don't exist yet"""
    if dst == '' or os.path.exists(dst):
        return
    head, _ = os.path.split(dst)
    if os.sep == ':' and not ':' in head:
        head = head + ':'
    mkdirs_thread_safe(head)
    try:
        os.mkdir(dst, 0o777)
    except OSError as err:
        if err.errno != 17:  # file exists
            raise


def make_sure_dir_exists(filename):
    """ Makes sure that the path to file exists, but creating directories. """
    dirname = os.path.dirname(filename)
    # dir == '' for current dir
    if dirname != '' and not os.path.exists(dirname):
        mkdirs_thread_safe(dirname)



from zuper_commons.text import pretty_msg

__all__ = [
    'check_isinstance',
]


def check_isinstance(ob, expected, **kwargs):
    if not isinstance(ob, expected):
        kwargs['object'] = ob
        raise_type_mismatch(ob, expected, **kwargs)


def raise_type_mismatch(ob, expected, **kwargs):
    """ Raises an exception concerning ob having the wrong type. """
    e = 'Object not of expected type:'
    e += '\n  expected: {}'.format(expected)
    e += '\n  obtained: {}'.format(type(ob))
    msg = pretty_msg(e, **kwargs)
    raise ValueError(msg)



# -*- coding: utf-8 -*-
import os

__all__ = [
    'dir_from_package_name',
]


def dir_from_package_name(d: str):
    """ This works for "package.sub" format. If it's only
        package, we look for __init__.py"""
    tokens = d.split('.')
    if len(tokens) == 1:
        package = d
        sub = '__init__'
    else:
        package = '.'.join(tokens[:-1])
        sub = tokens[-1]
    try:
        from pkg_resources import resource_filename  # @UnresolvedImport
        res = resource_filename(package, sub + '.py')

        if len(tokens) == 1:
            res = os.path.dirname(res)

        return res
    except BaseException as e:  # pragma: no cover
        msg = format_error('Cannot resolve package name', d=d)
        raise ValueError(msg) from e



import io
import logging
import traceback
from logging import currentframe

from os.path import normcase


def monkeypatch_findCaller():
    if __file__.lower()[-4:] in ['.pyc', '.pyo']:
        _wrapper_srcfile = __file__.lower()[:-4] + '.py'
    else:
        _wrapper_srcfile = __file__
    _wrapper_srcfile = normcase(_wrapper_srcfile)

    def findCaller(self, stack_info=False):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = currentframe()
        # On some versions of IronPython, currentframe() returns None if
        # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back
        rv = "(unknown file)", 0, "(unknown function)", None
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = normcase(co.co_filename)
            if filename == _wrapper_srcfile or filename == logging._srcfile:
                f = f.f_back
                continue
            sinfo = None
            if stack_info:
                sio = io.StringIO()
                sio.write('Stack (most recent call last):\n')
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                if sinfo[-1] == '\n':
                    sinfo = sinfo[:-1]
                sio.close()
            rv = (co.co_filename, f.f_lineno, co.co_name, sinfo)
            break
        return rv

    logging.Logger.findCaller = findCaller





from typing import *

__all__ = ['indent']


def indent(s: str, prefix: str, first: Optional[str] = None) -> str:
    if not isinstance(s, str):
        s = u'{}'.format(s)

    assert isinstance(prefix, str), type(prefix)
    try:
        lines = s.split('\n')
    except UnicodeDecodeError:
        print(type(s))  # XXX
        print(s)  # XXX
        lines = [s]
    if not lines:
        return u''

    if first is None:
        first = prefix

    m = max(len(prefix), len(first))

    prefix = ' ' * (m - len(prefix)) + prefix
    first = ' ' * (m - len(first)) + first

    # differnet first prefix
    res = [u'%s%s' % (prefix, line.rstrip()) for line in lines]
    res[0] = u'%s%s' % (first, lines[0].rstrip())
    return '\n'.join(res)



# coding=utf-8
import logging

import termcolor

__all__ = ['setup_logging_color', 'setup_logging_format', 'setup_logging']


def get_FORMAT_datefmt():
    pre = '%(asctime)s|%(name)s|%(filename)s:%(lineno)s|%(funcName)s(): '
    pre = termcolor.colored(pre, attrs=['dark'])
    FORMAT = pre + "%(message)s"
    datefmt = "%H:%M:%S"
    return FORMAT, datefmt

def setup_logging_format():
    from logging import Logger, StreamHandler, Formatter
    import logging

    FORMAT, datefmt  = get_FORMAT_datefmt()
    logging.basicConfig(format=FORMAT, datefmt=datefmt)

    if Logger.root.handlers:  # @UndefinedVariable
        for handler in Logger.root.handlers:  # @UndefinedVariable
            if isinstance(handler, StreamHandler):
                formatter = Formatter(FORMAT, datefmt=datefmt)
                handler.setFormatter(formatter)
    else:
        logging.basicConfig(format=FORMAT, datefmt=datefmt)


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if levelno >= 50:
            color = '\x1b[31m'  # red
        elif levelno >= 40:
            color = '\x1b[31m'  # red
        elif levelno >= 30:
            color = '\x1b[33m'  # yellow
        elif levelno >= 20:
            color = '\x1b[32m'  # green
        elif levelno >= 10:
            color = '\x1b[35m'  # pink
        else:
            color = '\x1b[0m'  # normal

        msg = str(args[1].msg)

        lines = msg.split('\n')

        def color_line(l):
            return "%s%s%s" % (color, l, '\x1b[0m')  # normal

        lines = list(map(color_line, lines))

        args[1].msg = "\n".join(lines)
        return fn(*args)

    return new


def setup_logging_color():
    import platform

    if platform.system() != 'Windows':
        emit2 = add_coloring_to_emit_ansi(logging.StreamHandler.emit)
        logging.StreamHandler.emit = emit2


def setup_logging():
    # logging.basicConfig()
    setup_logging_color()
    setup_logging_format()



__all__ = [
    'describe_type',
]

def describe_type(x):
    """ Returns a friendly description of the type of x. """

    if hasattr(x, '__class__'):
        c = x.__class__
        if hasattr(x, '__name__'):
            class_name = '%s' % c.__name__
        else:
            class_name = str(c)
    else:
        # for extension classes (spmatrix)
        class_name = str(type(x))

    return class_name



from typing import *

from .zc_indenting import indent

__all__ = [
    'pretty_dict',
    'pretty_msg',
    'format_error',
]


def pretty_msg(head: str, **kwargs: Any):
    return pretty_dict(head, kwargs)

format_error = pretty_msg

def pretty_dict(head: Optional[str],
                d: Dict[str, Any],
                omit_falsy=False,
                sort_keys=False):
    if not d:
        return head + ':  (empty dict)' if head else '(empty dict)'
    s = []
    n = max(len(str(_)) for _ in d)

    ordered = sorted(d) if sort_keys else list(d)
    # ks = sorted(d)
    for k in ordered:
        v = d[k]

        if k == '__builtins__':
            v = '(hiding __builtins__)'

        if not hasattr(v, 'conclusive') and (not isinstance(v, int)) and (not v) and omit_falsy:
            continue
        prefix = (str(k) + ':').rjust(n + 1) + ' '

        if isinstance(v, TypeVar):
            # noinspection PyUnresolvedReferences
            v = f'TypeVar({v.__name__}, bound={v.__bound__})'
        if isinstance(v, dict):
            v = pretty_dict('', v)
        s.append(indent(v, '', prefix))

    return (head + ':\n' if head else '') + indent("\n".join(s), "│ ")



# -*- coding: utf-8 -*-


import os

__all__ = [
    'friendly_path',
]


def friendly_path(path, use_environment=True):
    """ 
        Gets a friendly representation of the given path,
        using relative paths or environment variables
        (if use_environment = True).
    """
    # TODO: send extra rules

    options = []

    options.append(os.path.relpath(path, os.getcwd()))

    rules = []
    rules.append(('~', os.path.expanduser('~')))
    rules.append(('.', os.getcwd()))
    rules.append(('.', os.path.realpath(os.getcwd())))

    if use_environment:
        envs = dict(os.environ)
        # remove unwanted 
        for e in list(envs.keys()):
            if 'PWD' in e:
                del envs[e]

        for k, v in envs.items():
            if v:
                if v and v[-1] == '/':
                    v = v[:-1]
                if v and v[0] == '/':
                    rules.append(('${%s}' % k, v))

    # apply longest first
    rules.sort(key=lambda x: (-len(x[1])))
    path = replace_variables(path, rules)

    options.append(path)

    weight_doubledot = 5

    def score(s):
        # penalize '..' a lot
        s = s.replace('..', '*' * weight_doubledot)
        return len(s)

    options.sort(key=score)
    result = options[0]

    # print('Converted %s  => %s' % (original, result))

    return result


def replace_variables(path, rules):
    for k, v in rules:
        if path.startswith(v):
            # print("  applied %s => %s" % (v, k))
            path = path.replace(v, k)
    return path



# -*- coding: utf-8 -*-
import codecs
import os

from zuper_commons.types import check_isinstance
from . import logger
from .zc_friendly_path import friendly_path
from .zc_mkdirs import make_sure_dir_exists
from .zc_path_utils import expand_all

__all__ = [
    'read_bytes_from_file',
    'read_ustring_from_utf8_file',
    'read_ustring_from_utf8_file_lenient',
    'write_bytes_to_file',
    'write_ustring_to_utf8_file',
]


def read_bytes_from_file(filename: str) -> bytes:
    """ Read binary data and returns bytes """
    _check_exists(filename)
    with open(filename, 'rb') as f:
        return f.read()


def read_ustring_from_utf8_file(filename: str) -> str:
    """ Returns a unicode/proper string """
    _check_exists(filename)
    with codecs.open(filename, encoding='utf-8') as f:
        try:
            return f.read()
        except UnicodeDecodeError as e:
            msg = 'Could not successfully decode file %s' % filename
            raise UnicodeError(msg) from e


def read_ustring_from_utf8_file_lenient(filename) -> str:
    """ Ignores errors """
    _check_exists(filename)
    with codecs.open(filename, encoding='utf-8', errors='ignore') as f:
        try:
            return f.read()
        except UnicodeDecodeError as e:
            msg = 'Could not successfully decode file %s' % filename
            raise UnicodeError(msg) from e


def _check_exists(filename):
    if not os.path.exists(filename):
        if os.path.lexists(filename):
            msg = 'The link %s does not exist.' % filename
            msg += ' it links to %s' % os.readlink(filename)
            raise ValueError(msg)
        else:
            msg = 'Could not find file %r' % filename
            msg += ' from directory %s' % os.getcwd()
            raise ValueError(msg)


def write_ustring_to_utf8_file(data: str, filename, quiet=False):
    """
    It also creates the directory if it does not exist.
    :param data:
    :param filename:
    :param quiet:
    :return:
    """
    check_isinstance(data, str)
    b = data.encode('utf-8')  # OK
    return write_bytes_to_file(b, filename, quiet=quiet)


def write_bytes_to_file(data: bytes, filename: str, quiet=False):
    """
        Writes the data to the given filename.
        If the data did not change, the file is not touched.

    """
    check_isinstance(data, bytes)

    L = len(filename)
    if L > 1024:
        msg = f'Invalid argument filename: too long at {L}. Did you confuse it with data?\n{filename[:1024]}'
        raise ValueError(msg)

    filename = expand_all(filename)
    make_sure_dir_exists(filename)

    if os.path.exists(filename):
        current = open(filename, 'rb').read()
        if current == data:
            if not 'assets' in filename:
                if not quiet:
                    logger.debug('already up to date %s' % friendly_path(filename))
            return

    with open(filename, 'wb') as f:
        f.write(data)

    if filename.startswith('/tmp'):
        quiet = True

    if not quiet:
        size = '%.1fMB' % (len(data) / (1024 * 1024))
        logger.debug('Written %s to: %s' % (size, friendly_path(filename)))



# -*- coding: utf-8 -*-

import pickle
import traceback
from io import BytesIO
from pickle import (Pickler, SETITEM, MARK, SETITEMS, EMPTY_TUPLE, TUPLE, POP,
                    _tuplesize2code, POP_MARK)

from zuper_commons.types.zc_describe_type import describe_type
from . import logger

__all__ = ['find_pickling_error']


def find_pickling_error(obj, protocol=pickle.HIGHEST_PROTOCOL):
    sio = BytesIO()
    try:
        pickle.dumps(obj)
    except BaseException:
        se1 = traceback.format_exc()

        pickler = MyPickler(sio, protocol)
        try:
            pickler.dump(obj)
        except Exception:
            se2 = traceback.format_exc()
            msg = pickler.get_stack_description()
            msg += '\n --- Current exception----\n%s' % se1
            msg += '\n --- Old exception----\n%s' % se2
            return msg
        else:
            msg = 'I could not find the exact pickling error.'
            raise Exception(msg)
    else:
        msg = ('Strange! I could not reproduce the pickling error '
               'for the object of class %s' % describe_type(obj))
        logger.info(msg)


class MyPickler(Pickler):
    def __init__(self, *args, **kargs):
        Pickler.__init__(self, *args, **kargs)
        self.stack = []

    def save(self, obj):
        desc = 'object of type %s' % (describe_type(obj))
        # , describe_value(obj, 100))
        #  self.stack.append(describe_value(obj, 120))
        self.stack.append(desc)
        Pickler.save(self, obj)
        self.stack.pop()

    def get_stack_description(self):
        s = 'Pickling error occurred at:\n'
        for i, context in enumerate(self.stack):
            s += ' ' * i + '- %s\n' % context
        return s

    def save_pair(self, k, v):
        self.stack.append('key %r = object of type %s' % (k, describe_type(v)))
        self.save(k)
        self.save(v)
        self.stack.pop()

    def _batch_setitems(self, items):

        # Helper to batch up SETITEMS sequences; proto >= 1 only
        # save = self.save
        write = self.write

        if not self.bin:
            for k, v in items:
                self.stack.append('entry %s' % str(k))
                self.save_pair(k, v)
                self.stack.pop()
                write(SETITEM)
            return

        r = list(range(self._BATCHSIZE))
        while items is not None:
            tmp = []
            for _ in r:
                try:
                    tmp.append(items.next())
                except StopIteration:
                    items = None
                    break
            n = len(tmp)
            if n > 1:
                write(MARK)
                for k, v in tmp:
                    self.stack.append('entry %s' % str(k))
                    self.save_pair(k, v)
                    self.stack.pop()
                write(SETITEMS)
            elif n:
                k, v = tmp[0]
                self.stack.append('entry %s' % str(k))
                self.save_pair(k, v)
                self.stack.pop()
                write(SETITEM)
            # else tmp is empty, and we're done

    def save_tuple(self, obj):
        write = self.write
        proto = self.proto

        n = len(obj)
        if n == 0:
            if proto:
                write(EMPTY_TUPLE)
            else:
                write(MARK + TUPLE)
            return

        save = self.save
        memo = self.memo
        if n <= 3 and proto >= 2:
            for i, element in enumerate(obj):
                self.stack.append('tuple element %s' % i)
                save(element)
                self.stack.pop()
            # Subtle.  Same as in the big comment below.
            if id(obj) in memo:
                get = self.get(memo[id(obj)][0])
                write(POP * n + get)
            else:
                write(_tuplesize2code[n])
                self.memoize(obj)
            return

        # proto 0 or proto 1 and tuple isn't empty, or proto > 1 and tuple
        # has more than 3 elements.
        write(MARK)
        for i, element in enumerate(obj):
            self.stack.append('tuple element %s' % i)
            save(element)
            self.stack.pop()

        if id(obj) in memo:
            # Subtle.  d was not in memo when we entered save_tuple(), so
            # the process of saving the tuple's elements must have saved
            # the tuple itself:  the tuple is recursive.  The proper action
            # now is to throw away everything we put on the stack, and
            # simply GET the tuple (it's already constructed).  This check
            # could have been done in the "for element" loop instead, but
            # recursive tuples are a rare thing.
            get = self.get(memo[id(obj)][0])
            if proto:
                write(POP_MARK + get)
            else:  # proto 0 -- POP_MARK not available
                write(POP * (n + 1) + get)
            return

        # No recursion.
        self.write(TUPLE)
        self.memoize(obj)



# -*- coding: utf-8 -*-
import math

__all__ = [

    'duration_compact',
]


def duration_compact(seconds):
    seconds = int(math.ceil(seconds))
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    days, hours = divmod(hours, 24)
    years, days = divmod(days, 365)

    minutes = int(minutes)
    hours = int(hours)
    days = int(days)
    years = int(years)

    duration = []
    if years > 0:
        duration.append('%dy' % years)
    else:
        if days > 0:
            duration.append('%dd' % days)
        if (days < 3) and (years == 0):
            if hours > 0:
                duration.append('%dh' % hours)
            if (hours < 3) and (days == 0):
                if minutes > 0:
                    duration.append('%dm' % minutes)
                if (minutes < 3) and (hours == 0):
                    if seconds > 0:
                        duration.append('%ds' % seconds)

    return ' '.join(duration)

#
# def duration_human(seconds):
# ''' Code modified from
#     http://darklaunch.com/2009/10/06
#     /python-time-duration-human-friendly-timestamp
#     '''
#     seconds = int(math.ceil(seconds))
#     minutes, seconds = divmod(seconds, 60)
#     hours, minutes = divmod(minutes, 60)
#     days, hours = divmod(hours, 24)
#     years, days = divmod(days, 365.242199)
# 
#     minutes = int(minutes)
#     hours = int(hours)
#     days = int(days)
#     years = int(years)
# 
#     duration = []
#     if years > 0:
#         duration.append('%d year' % years + 's' * (years != 1))
#     else:
#         if days > 0:
#             duration.append('%d day' % days + 's' * (days != 1))
#         if (days < 3) and (years == 0):
#             if hours > 0:
#                 duration.append('%d hour' % hours + 's' * (hours != 1))
#             if (hours < 3) and (days == 0):
#                 if minutes > 0:
#                     duration.append('%d min' % minutes +
#                                      's' * (minutes != 1))
#                 if (minutes < 3) and (hours == 0):
#                     if seconds > 0:
#                         duration.append('%d sec' % seconds +
#                                          's' * (seconds != 1))
# 
#     return ' '.join(duration)




from setuptools import setup, find_packages


def get_version(filename):
    import ast
    version = None
    with open(filename) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            raise ValueError('No version found in %r.' % filename)
    if version is None:
        raise ValueError(filename)
    return version


version = get_version('src/zuper_commons/__init__.py')
import os

description = """"""


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='zuper-commons',
      version=version,
      package_dir={'': 'src'},
      packages=find_packages('src'),

      zip_safe=True,
      entry_points={
          'console_scripts': [

          ]
      },
      install_requires=[
          'PyContracts',
      ],
      )



# -*- coding: utf-8 -*-
import fnmatch
import os
import time
from typing import *
from collections import defaultdict

from contracts import contract
from contracts.utils import check_isinstance


from . import logger

__all__ = [
    'locate_files',
]


@contract(returns='list(str)', directory='str',
          pattern='str|seq(str)', followlinks='bool')
def locate_files(directory, pattern, followlinks=True,
                 include_directories=False,
                 include_files=True,
                 normalize=True,
                 ignore_patterns: Optional[List[str]]=None):
    """
        pattern is either a string or a sequence of strings

        NOTE: if you do not pass ignore_patterns, it will use  MCDPConstants.locate_files_ignore_patterns

        ignore_patterns = ['*.bak']

        normalize = uses realpath
    """
    t0 = time.time()

    if ignore_patterns is None:
        ignore_patterns = []

    if isinstance(pattern, str):
        patterns = [pattern]
    else:
        patterns = list(pattern)
        for p in patterns:
            check_isinstance(p, str)

    # directories visited
    # visited = set()
    # visited_basename = set()
    # print('locate_files %r %r' % (directory, pattern))
    filenames = []

    def matches_pattern(x):
        return any(fnmatch.fnmatch(x, _) or (x == _) for _ in patterns)

    def should_ignore_resource(x):
        return any(fnmatch.fnmatch(x, _) or (x == _) for _ in ignore_patterns)

    def accept_dirname_to_go_inside(_root_, d_):
        if should_ignore_resource(d_):
            return False
        # XXX
        # dd = os.path.realpath(os.path.join(root_, d_))
        # if dd in visited:
        #     return False
        # visited.add(dd)
        return True

    def accept_dirname_as_match(_):
        return include_directories and \
               not should_ignore_resource(_) and \
               matches_pattern(_)

    def accept_filename_as_match(_):
        return include_files and \
               not should_ignore_resource(_) and \
               matches_pattern(_)

    ntraversed = 0
    for root, dirnames, files in os.walk(directory, followlinks=followlinks):
        ntraversed += 1
        dirnames[:] = [_ for _ in dirnames if accept_dirname_to_go_inside(root, _)]
        for f in files:
            # logger.info('look ' + root + '/' + f)
            if accept_filename_as_match(f):
                filename = os.path.join(root, f)
                filenames.append(filename)
        for d in dirnames:
            if accept_dirname_as_match(d):
                filename = os.path.join(root, d)
                filenames.append(filename)

    if normalize:
        real2norm = defaultdict(lambda: [])
        for norm in filenames:
            real = os.path.realpath(norm)
            real2norm[real].append(norm)
            # print('%s -> %s' % (real, norm))

        for k, v in real2norm.items():
            if len(v) > 1:
                msg = 'In directory:\n\t%s\n' % directory
                msg += 'I found %d paths that refer to the same file:\n' % len(v)
                for n in v:
                    msg += '\t%s\n' % n
                msg += 'refer to the same file:\n\t%s\n' % k
                msg += 'I will silently eliminate redundancies.'
                # logger.warning(msg) # XXX

        filenames = list(real2norm.keys())

    seconds = time.time() - t0
    if seconds > 5:
        n = len(filenames)
        nuniques = len(set(filenames))
        logger.debug('%.4f s for locate_files(%s,%s): %d traversed, found %d filenames (%d uniques)' %
                     (seconds, directory, pattern, ntraversed, n, nuniques))
    return filenames



# -*- coding: utf-8 -*-
import gzip
import os
import random
from contextlib import contextmanager

__all__ = [
    'safe_write',
    'safe_read',
]


def is_gzip_filename(filename):
    return '.gz' in filename


@contextmanager
def safe_write(filename, mode='wb', compresslevel=5):
    """ 
        Makes atomic writes by writing to a temp filename. 
        Also if the filename ends in ".gz", writes to a compressed stream.
        Yields a file descriptor.
        
        It is thread safe because it renames the file.
        If there is an error, the file will be removed if it exists.
    """
    dirname = os.path.dirname(filename)
    if dirname:
        if not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except:
                pass

                # Dont do this!
                # if os.path.exists(filename):
                # os.unlink(filename)
                #     assert not os.path.exists(filename)
                #
    n = random.randint(0, 10000)
    tmp_filename = '%s.tmp.%s.%s' % (filename, os.getpid(), n)
    try:
        if is_gzip_filename(filename):
            fopen = lambda fname, fmode: gzip.open(filename=fname, mode=fmode,
                                                   compresslevel=compresslevel)
        else:
            fopen = open

        with fopen(tmp_filename, mode) as f:
            yield f
        f.close()

        # if os.path.exists(filename):
        # msg = 'Race condition for writing to %r.' % filename
        #             raise Exception(msg)
        #
        # On Unix, if dst exists and is a file, it will be replaced silently
        #  if the user has permission.
        os.rename(tmp_filename, filename)
    except:
        if os.path.exists(tmp_filename):
            os.unlink(tmp_filename)
        if os.path.exists(filename):
            os.unlink(filename)
        raise


@contextmanager
def safe_read(filename, mode='rb'):
    """ 
        If the filename ends in ".gz", reads from a compressed stream.
        Yields a file descriptor.
    """
    try:
        if is_gzip_filename(filename):
            f = gzip.open(filename, mode)
            try:
                yield f
            finally:
                f.close()

        else:
            with open(filename, mode) as f:
                yield f
    except:
        # TODO
        raise




def describe_value(x, clip=80):
    """ Describes an object, for use in the error messages.
        Short description, no multiline.
    """
    if hasattr(x, 'shape') and hasattr(x, 'dtype'):
        shape_desc = 'x'.join(str(i) for i in x.shape)
        desc = 'array[%r](%s) ' % (shape_desc, x.dtype)
        final = desc + clipped_repr(x, clip - len(desc))
        return remove_newlines(final)
    else:
        from .zc_describe_type import  describe_type
        class_name = describe_type(x)
        desc = 'Instance of %s: ' % class_name
        final = desc + clipped_repr(x, clip - len(desc))
        return remove_newlines(final)

def clipped_repr(x, clip):
    s = "{0!r}".format(x)
    if len(s) > clip:
        clip_tag = '... [clip]'
        cut = clip - len(clip_tag)
        s = "%s%s" % (s[:cut], clip_tag)
    return s

def remove_newlines(s):
    return s.replace('\n', ' ')




from typing import Union


__all__ = [
    'get_md5',
    'get_sha1',
]

def get_md5(contents: Union[bytes, str]) -> str:
    """ Returns an hexdigest (string).
        If the contents is a string, then it is encoded as utf-8.
    """
    if isinstance(contents, str):
        contents = contents.encode('utf-8')
    from zuper_commons.types import check_isinstance
    check_isinstance(contents, bytes)
    import hashlib
    m = hashlib.md5()
    m.update(contents)
    s = m.hexdigest()
    check_isinstance(s, str)
    return s


def get_sha1(contents: bytes) -> str:
    """ Returns an hexdigest (string) """
    from zuper_commons.types import check_isinstance
    import hashlib
    check_isinstance(contents, bytes)
    m = hashlib.sha1()
    m.update(contents)
    s = m.hexdigest()
    check_isinstance(s, str)
    return s



from .. import logger
logger = logger.getChild("fs")

from .zc_dir_from_package_nam import *
from .zc_fileutils import *
from .zc_locate_files_imp import *
from .zc_mkdirs import *
from .zc_safe_pickling import *
from .zc_path_utils import *
from .zc_safe_write import *
from .zc_friendly_path import *
from .zc_debug_pickler import *
